@@ -232,11 +232,13 @@ on_agent_props_changed (GDBusProxy *agent_proxy,
         g_variant_get (changed_properties, "a{sv}", &iter);
         while (g_variant_iter_loop (iter, "{&sv}", &key, &value)) {
                 GClueAccuracyLevel max_accuracy;
+                GClueConfig *config;
                 const char *id;
 
                 if (strcmp (key, "MaxAccuracyLevel") != 0)
                         continue;
 
+                config = gclue_config_get_singleton ();
                 id = gclue_client_get_desktop_id (GCLUE_CLIENT (client));
                 max_accuracy = g_variant_get_uint32 (value);
                 /* FIXME: We should be handling all values of max accuracy
@@ -253,7 +255,8 @@ on_agent_props_changed (GDBusProxy *agent_proxy,
                         start_client (client, accuracy);
                         g_debug ("Re-started '%s'.", id);
                 } else if (max_accuracy == 0 &&
-                           gclue_client_get_active (GCLUE_CLIENT (client))) {
+                           gclue_client_get_active (GCLUE_CLIENT (client)) &&
+                           !gclue_config_is_system_component (config, id)) {
                         stop_client (client);
                         client->priv->agent_stopped = TRUE;
                         g_debug ("Stopped '%s'.", id);
@@ -335,6 +338,8 @@ gclue_service_client_handle_start (GClueClient           *client,
         StartData *data;
         const char *desktop_id;
         GClueAccuracyLevel accuracy_level, max_accuracy;
+        GClueAppPerm app_perm;
+        guint32 uid;
 
         if (priv->locator != NULL)
                 /* Already started */
@@ -349,18 +354,32 @@ gclue_service_client_handle_start (GClueClient           *client,
                 return TRUE;
         }
 
+        config = gclue_config_get_singleton ();
+        uid = gclue_client_info_get_user_id (priv->client_info);
+        app_perm = gclue_config_get_app_perm (config,
+                                              desktop_id,
+                                              priv->client_info);
+        if (app_perm == GCLUE_APP_PERM_DISALLOWED) {
+                g_dbus_method_invocation_return_error (invocation,
+                                                       G_DBUS_ERROR,
+                                                       G_DBUS_ERROR_ACCESS_DENIED,
+                                                       "'%s' disallowed by "
+                                                       "configuration for UID %u",
+                                                       desktop_id,
+                                                       uid);
+                return TRUE;
+        }
+
         data = g_slice_new (StartData);
         data->client = g_object_ref (client);
         data->invocation =  g_object_ref (invocation);
 
-        config = gclue_config_get_singleton ();
         accuracy_level = gclue_client_get_requested_accuracy_level (client);
 
         /* No agent == No authorization needed */
         if (priv->agent_proxy == NULL ||
-            gclue_config_is_app_allowed (config,
-                                         desktop_id,
-                                         priv->client_info)) {
+            gclue_config_is_system_component (config, desktop_id) ||
+            app_perm == GCLUE_APP_PERM_ALLOWED) {
                 complete_start (data, accuracy_level);
 
                 return TRUE;
@@ -368,8 +387,6 @@ gclue_service_client_handle_start (GClueClient           *client,
 
         max_accuracy = gclue_agent_get_max_accuracy_level (priv->agent_proxy);
         if (max_accuracy == 0) {
-                guint32 uid = gclue_client_info_get_user_id (priv->client_info);
-
                 g_dbus_method_invocation_return_error (invocation,
                                                        G_DBUS_ERROR,
                                                        G_DBUS_ERROR_ACCESS_DENIED,
@@ -379,6 +396,9 @@ gclue_service_client_handle_start (GClueClient           *client,
                 start_data_free (data);
                 return TRUE;
         }
+        g_debug ("requested accuracy level: %u. "
+                 "Accuracy level allowed by agent: %u",
+                 accuracy_level, max_accuracy);
         accuracy_level = CLAMP (accuracy_level, 0, max_accuracy);
 
         gclue_agent_call_authorize_app (priv->agent_proxy,
@@ -409,6 +429,9 @@ gclue_service_client_finalize (GObject *object)
 
         g_clear_pointer (&priv->path, g_free);
         g_clear_object (&priv->connection);
+        g_signal_handlers_disconnect_by_func (priv->agent_proxy,
+                                              G_CALLBACK (on_agent_props_changed),
+                                              object);
         g_clear_object (&priv->agent_proxy);
         g_clear_object (&priv->locator);
         g_clear_object (&priv->location);
